
/*
   SEE WARNING BELOW!

   Builtin functions are designed to behave like interpreted functions.
   There are methods Function.operate() and Function.execute() which
   parallel the methods Builtin.operate() and Builtin.execute() below.
   Both kind of functions are not loaded until invoked by the user.
   Interpreted functions are defined in .m text files on the disk
   and when invoked are compiled by Function.compile; the interpreter
   code is kept in the table Global.functions. Builtin functions are
   contained in container classes which extend Builtin. They are
   constructed when first invoked by the user and stored in the array
   Builtin.functions[] below. If invoked a second time, the corresponding
   entry in the array will be not null and a new instance is not
   constructed. Each container class overrides the operate and execute
   methods it implements.

   Arguments in Minimat are passed by value: in a call B=f(A)
   the function f may change its input without fear of changing
   the value of the matrix A. This is to hold for both Builtin
   functions and interpreted functions. For calls via operate() to
   builtin functions,the responsibility for making the copy is left
   to the interpreter (see case OPn:); other calls pass the execution
   stack and the copy is made by execute() (for builtin functions)
   or bind() (for interpreted functions)

   The names of the builtin functions are kept in the array name[]
   below. Each sub class overrides one or  more of the four functions
   operate() ... operate(Matrix,Matrix,Matrix).    The k-th function
   overides those versions of operate which accept between minArgsin[k]
   and maxArgsin[k] arguments. An error is generated unless the function
   invocation satisfies

     minArgsin[k] <= argsin <= maxArgsin[k] &&
     minArgsout[k] <= argsout <= maxArgsout[k]

   The naming convention for the container classes is as follows:
   Functions which are performed elementwise end in 'E' or 'N';
   the 'N' indicates a function which is `Numeric' and not defined
   over fields more general than the complex numbers. The other
   classes which are handled by BuildBuiltin end in 'M'. Those
   which do not end in a capital letter are defined in
   BuiltinTail.java.

   To facilitate modifying the program, parts of Builtin.java
   are automatically generated by the program BuildBuiltin.
   This prevents errors in the correspondence between the six
   arrays name[],...,function[k]. The first part of Builtin.java is
   copied by BuildBuiltin from the file BuiltinHead.java.  BuildBuiltin
   generates the middle part and then copies the rest from the file
   BuiltinTail.java.

   WARNING: The current version of BuildBuiltin overwrites Builtin.java.


   ToDo: Most of the operations are performed elementwise so that the
   operate method consists of a double loop with a scalar method
   invoked in the body. One could save a bit of space by making the
   the double loop the default and invoking an operate method in
   a container class for the scalar method.
*/


import java.lang.Math;
import java.util.*;

  class Builtin{


    private static final int NOTFOUND = Global.NOTFOUND;

    static boolean find(String id){
        return indexOf(id)!= NOTFOUND;
    }

    private static int indexOf(String id){
        for(int i=0; i<name.length; i++){
            if(id.equals(name[i])) return i;
        }
        return NOTFOUND;
    }

    static Builtin getBuiltin(String id, int nin, int nout)
        throws MinimatException{
        int index = indexOf(id);
        if(index == NOTFOUND) error("Builtin Function '"+id+"' not found");
        if(nin<minArgsin[index]) error("Too few inputs for builtin Function '"+id+"'");
        if(maxArgsin[index]<nin) error("Too many inputs for builtin Function '"+id+"'");
        if(nout<minArgsout[index]) error("Too few outputs for builtin Function '"+id+"'");
        if(maxArgsout[index]<nout) error("Too many outputs for builtin Function '"+id+"'");
        if(functions[index]==null) create(index);
        assert(functions[index]!=null, "getBuiltIn");
        return functions[index];
    }


    String nameOf(){
        for(int i=0; i<functions.length; i++){
            if(this==functions[i])
                return name[i];
        }
        return "Unknown";
    }

    static String dump(Builtin b){
        for(int i=0; i<functions.length; i++){
            if(b==functions[i])
                return name[i];
        }
        return "unknown";
    }


    static String dump(){
        String str="***** Active Buitin Functions *********\n";
        for(int i=0; i<functions.length; i++){
            if(functions[i]!=null)
                str += name[i]+" " + functions[i]+"\n";
        }
        return str;
    }

    private static String unimplemented= "unimplemented function: ";

    Matrix operate()
        throws MinimatException{
        throw new MinimatException(unimplemented+nameOf());
    }

    Matrix operate(Matrix A)
        throws MinimatException{
        throw new MinimatException(unimplemented+nameOf());
    }

    Matrix operate(Matrix A, Matrix B)
        throws MinimatException{
        throw new MinimatException(unimplemented+nameOf());
    }

    Matrix operate(Matrix A, Matrix B, Matrix C)
        throws MinimatException{
        throw new MinimatException(unimplemented+nameOf());
    }


    /* The default method handles stupid invocations like
       [] = f(..) or [Y]=f(...).
    */
    Stack execute(Stack stack,int nargsin, int nargsout)
        throws MinimatException{
        Matrix A=null;Matrix B=null;Matrix C=null;
        Matrix Y;
        switch(nargsin){
            case 0:
                Y= operate();
            break;
            case 1:
                A=new Matrix((Matrix)stack.pop());
                Y= operate(A);
            break;
            case 2:
                B=new Matrix((Matrix)stack.pop());
                A=new Matrix((Matrix)stack.pop());
                Y= operate(A,B);
            break;
            case 3:
                C=new Matrix((Matrix)stack.pop());
                B=new Matrix((Matrix)stack.pop());
                A=new Matrix((Matrix)stack.pop());
                Y= operate(A,B,C);
            break;
            default:throw new MinimatException("unimplemented function");
        }
        Stack result = new Stack();
        if(nargsout == 1) result.push(Y);
        else
        if(nargsout!=0)
        throw new MinimatException("unimplemented function");
        return result;
    }

    static boolean insertBrackets(String id){
    // return true if id is name of a function (like diag)
    // which can be called in either of the forms
    //   id([a b ... c]) or id(a b ... c).

	return
	    id.equalsIgnoreCase("diag") ||
		id.equalsIgnoreCase("max")  ||
		id.equalsIgnoreCase("min")  ||
		id.equalsIgnoreCase("any")  ||
		id.equalsIgnoreCase("all")  ||
		id.equalsIgnoreCase("sum");
    }


    private static void assert(boolean condition, String where){
        Global.assert(condition, "Builtin.java (or BuiltinHead.java): "+where);
    }


     static void error(String msg) throws MinimatException{
        throw new MinimatException(msg);
    }


// *** Add text generated by BuildBuiltin.java



// *** Begin text generated by BuildBuiltin.java

// *** WARNING: ":", ",", ";" used in Function

 static String[] name = {
      ",",     ";",     ":",     "+",     "-",     "*",     "/",     
      "inv",   "'",     ".'",    "==",    "~=",    "<>",    ">",     
      ">=",    "<",     "<=",    "&",     "|",     "~",     "^",     
      ".*",    "./",    ".^",    "{}",    "rand",  "eye",   "zeros", 
      "ones",  "size",  "any",   "all",   "sum",   "min",   "max",   
      "diag",  "short", "shorte","long",  "longe", "pi",    "abs",   
      "acos",  "asin",  "atan",  "conj",  "cos",   "exp",   "fix",   
      "imag",  "log",   "real",  "round", "sign",  "sin",   "sqrt",  
      "tan",   "atan2", "expm",  "lu",    "det",   "eig",   "svd",   
      "qr"
      };

 static int[] minArgsin = {
      2,       2,       2,       2,       1,       2,       2,       
      1,       1,       1,       2,       2,       2,       2,       
      2,       2,       2,       2,       2,       1,       2,       
      2,       2,       2,       2,       0,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       0,       0,       0,       0,       0,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       2,       1,       1,       1,       1,       1,       
      1
      };

 static int[] maxArgsin = {
      2,       2,       3,       2,       2,       2,       2,       
      1,       1,       1,       2,       2,       2,       2,       
      2,       2,       2,       2,       2,       1,       2,       
      2,       2,       2,       2,       2,       1,       2,       
      2,       1,       1,       1,       1,       1,       1,       
      2,       0,       0,       0,       0,       0,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       2,       1,       1,       1,       1,       1,       
      1
      };

 static int[] minArgsout = {
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1
      };

 static int[] maxArgsout = {
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       2,       1,       1,       1,       2,       2,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       1,       1,       1,       1,       
      1,       1,       1,       2,       1,       2,       3,       
      2
      };

 static Builtin[] functions = {
      null,    null,    null,    null,    null,    null,    null,    
      null,    null,    null,    null,    null,    null,    null,    
      null,    null,    null,    null,    null,    null,    null,    
      null,    null,    null,    null,    null,    null,    null,    
      null,    null,    null,    null,    null,    null,    null,    
      null,    null,    null,    null,    null,    null,    null,    
      null,    null,    null,    null,    null,    null,    null,    
      null,    null,    null,    null,    null,    null,    null,    
      null,    null,    null,    null,    null,    null,    null,    
      null
      };

 private static void create(int index){
        switch(index){
         case 0: functions[0] = new ColcatM();break;
         case 1: functions[1] = new RowcatM();break;
         case 2: functions[2] = new RangeM();break;
         case 3: functions[3] = new AddE();break;
         case 4: functions[4] = new SubE();break;
         case 5: functions[5] = new MulM();break;
         case 6: functions[6] = new DivM();break;
         case 7: functions[7] = new InvM();break;
         case 8: functions[8] = new ConjposeM();break;
         case 9: functions[9] = new TransposeM();break;
         case 10: functions[10] = new EquE();break;
         case 11: functions[11] = new NeqE();break;
         case 12: functions[12] = new NeqE();break;
         case 13: functions[13] = new GtrE();break;
         case 14: functions[14] = new GeqE();break;
         case 15: functions[15] = new LesE();break;
         case 16: functions[16] = new LeqE();break;
         case 17: functions[17] = new AndE();break;
         case 18: functions[18] = new OrE();break;
         case 19: functions[19] = new NotE();break;
         case 20: functions[20] = new PowerM();break;
         case 21: functions[21] = new MulE();break;
         case 22: functions[22] = new DivE();break;
         case 23: functions[23] = new PowerE();break;
         case 24: functions[24] = new CommutatorM();break;
         case 25: functions[25] = new RandM();break;
         case 26: functions[26] = new EyeM();break;
         case 27: functions[27] = new ZerosM();break;
         case 28: functions[28] = new OnesM();break;
         case 29: functions[29] = new Size();break;
         case 30: functions[30] = new AnyM();break;
         case 31: functions[31] = new AllM();break;
         case 32: functions[32] = new SumM();break;
         case 33: functions[33] = new Min();break;
         case 34: functions[34] = new Max();break;
         case 35: functions[35] = new DiagM();break;
         case 36: functions[36] = new ShortF();break;
         case 37: functions[37] = new ShorteF();break;
         case 38: functions[38] = new LongF();break;
         case 39: functions[39] = new LongeF();break;
         case 40: functions[40] = new PiN();break;
         case 41: functions[41] = new AbsN();break;
         case 42: functions[42] = new AcosN();break;
         case 43: functions[43] = new AsinN();break;
         case 44: functions[44] = new AtanN();break;
         case 45: functions[45] = new ConjN();break;
         case 46: functions[46] = new CosN();break;
         case 47: functions[47] = new ExpN();break;
         case 48: functions[48] = new FixN();break;
         case 49: functions[49] = new ImagN();break;
         case 50: functions[50] = new LogN();break;
         case 51: functions[51] = new RealN();break;
         case 52: functions[52] = new RoundN();break;
         case 53: functions[53] = new SignN();break;
         case 54: functions[54] = new SinN();break;
         case 55: functions[55] = new SqrtN();break;
         case 56: functions[56] = new TanN();break;
         case 57: functions[57] = new Atan2N();break;
         case 58: functions[58] = new Exp();break;
         case 59: functions[59] = new LUd();break;
         case 60: functions[60] = new Det();break;
         case 61: functions[61] = new Eig();break;
         case 62: functions[62] = new SVd();break;
         case 63: functions[63] = new QRd();break;
        }

      }//create
}//Builtin
class ColcatM extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         return Matrix.colcat(A,B);
      }//operate
}//ColcatM

class RowcatM extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         return Matrix.rowcat(A,B);
      }//operate
}//RowcatM

class RangeM extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         return Matrix.range(A,B);
      }//operate

      Matrix operate(Matrix A, Matrix B,Matrix C)
           throws MinimatException{
         return Matrix.range(A,B,C);
      }//operate
}//RangeM

class AddE extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,"+");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.add(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//AddE

class SubE extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.subtract(A.entry[i][j]);
         return B;
      }//operate

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,"-");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.subtract(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//SubE

class MulM extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         return Matrix.multiply(A,B);
      }//operate
}//MulM

class DivM extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         return Matrix.divide(A,B);
      }//operate
}//DivM

class InvM extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         return Matrix.inverse(A);
      }//operate
}//InvM

class ConjposeM extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         return Matrix.conjpose(A);
      }//operate
}//ConjposeM

class TransposeM extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         return Matrix.transpose(A);
      }//operate
}//TransposeM

class EquE extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,"==");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.equ(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//EquE

class NeqE extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,"~=");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.neq(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//NeqE

class GtrE extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,">");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.gtr(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//GtrE

class GeqE extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,">=");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.geq(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//GeqE

class LesE extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,"<");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.les(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//LesE

class LeqE extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,"<=");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.leq(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//LeqE

class AndE extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,"&");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.and(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//AndE

class OrE extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,"|");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.or(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//OrE

class NotE extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.not(A.entry[i][j]);
         return B;
      }//operate
}//NotE

class PowerM extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         return Matrix.power(A,B);
      }//operate
}//PowerM

class MulE extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,".*");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.multiply(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//MulE

class DivE extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,"./");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.divide(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//DivE

class PowerE extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,".^");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.power(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//PowerE

class CommutatorM extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         return Matrix.commutator(A,B);
      }//operate
}//CommutatorM

class RandM extends Builtin{

      Matrix operate()
           throws MinimatException{
         return Matrix.rand();
      }//operate

      Matrix operate(Matrix A)
           throws MinimatException{
         return Matrix.rand(A);
      }//operate

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         return Matrix.rand(A,B);
      }//operate
}//RandM

class EyeM extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         return Matrix.eye(A);
      }//operate
}//EyeM

class ZerosM extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         return Matrix.zeros(A);
      }//operate

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         return Matrix.zeros(A,B);
      }//operate
}//ZerosM

class OnesM extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         return Matrix.ones(A);
      }//operate

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         return Matrix.ones(A,B);
      }//operate
}//OnesM

class AnyM extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         return Matrix.any(A);
      }//operate
}//AnyM

class AllM extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         return Matrix.all(A);
      }//operate
}//AllM

class SumM extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         return Matrix.sum(A);
      }//operate
}//SumM

class DiagM extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         return Matrix.diag(A);
      }//operate

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         return Matrix.diag(A,B);
      }//operate
}//DiagM

class PiN extends Builtin{

      Matrix operate()
           throws MinimatException{
         Matrix A = new Matrix(Scalar.pi());
         return A;
      }//operate
}//PiN

class AbsN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.abs(A.entry[i][j]);
         return B;
      }//operate
}//AbsN

class AcosN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.acos(A.entry[i][j]);
         return B;
      }//operate
}//AcosN

class AsinN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.asin(A.entry[i][j]);
         return B;
      }//operate
}//AsinN

class AtanN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.atan(A.entry[i][j]);
         return B;
      }//operate
}//AtanN

class ConjN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.conj(A.entry[i][j]);
         return B;
      }//operate
}//ConjN

class CosN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.cos(A.entry[i][j]);
         return B;
      }//operate
}//CosN

class ExpN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.exp(A.entry[i][j]);
         return B;
      }//operate
}//ExpN

class FixN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.fix(A.entry[i][j]);
         return B;
      }//operate
}//FixN

class ImagN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.imag(A.entry[i][j]);
         return B;
      }//operate
}//ImagN

class LogN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.log(A.entry[i][j]);
         return B;
      }//operate
}//LogN

class RealN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.real(A.entry[i][j]);
         return B;
      }//operate
}//RealN

class RoundN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.round(A.entry[i][j]);
         return B;
      }//operate
}//RoundN

class SignN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.sign(A.entry[i][j]);
         return B;
      }//operate
}//SignN

class SinN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.sin(A.entry[i][j]);
         return B;
      }//operate
}//SinN

class SqrtN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.sqrt(A.entry[i][j]);
         return B;
      }//operate
}//SqrtN

class TanN extends Builtin{

      Matrix operate(Matrix A)
           throws MinimatException{
         Matrix B = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             B.entry[i][j] = Scalar.tan(A.entry[i][j]);
         return B;
      }//operate
}//TanN

class Atan2N extends Builtin{

      Matrix operate(Matrix A, Matrix B)
           throws MinimatException{
         Matrix.sameSize(A,B,"atan2");
         Matrix C = new Matrix(A.nrows,A.ncols);
         for(int i=0;i<A.nrows;i++)
           for(int j=0;j<A.ncols;j++)
             C.entry[i][j] = Scalar.atan2(A.entry[i][j], B.entry[i][j]);
         return C;
      }//operate
}//Atan2N

// *** End of text generated by BuildBuiltin


// Following code is not generated by BuildBuiltin
// It is in the file BuiltinTail.java and catenated
// onto the end of Builtin.java by BuildBuiltin

class Size extends Builtin{
    Matrix operate(Matrix A)throws MinimatException{
        Matrix S = new Matrix(1,2);
        S.entry[0][0]=new Scalar(A.nrows);
        S.entry[0][1]=new Scalar(A.ncols);
        return S;
    }
    Stack execute(Stack stack,int nargsin, int nargsout)
      throws MinimatException{
        if(nargsout!=2 || nargsin!=1)
            return super.execute(stack,nargsin,nargsout);
        else{
            Matrix A = (Matrix) stack.pop();
            Matrix m = new Matrix(new Scalar(A.nrows));
            Matrix n = new Matrix(new Scalar(A.ncols));
            Stack result = new Stack();
            result.push(m);
            result.push(n);
            return result;
        }
      }
}
class ShortF extends Builtin{}
class ShorteF extends Builtin{}
class LongF extends Builtin{}
class LongeF extends Builtin{}

class Min extends Builtin{
    Matrix operate(Matrix A)throws MinimatException{
        return Matrix.min(A);
    }
    Stack execute(Stack stack,int nargsin, int nargsout)
      throws MinimatException{
        if(nargsout!=2)
            return super.execute(stack,nargsin,nargsout);
        else{
            Matrix A = (Matrix)stack.pop();
            Matrix R = Matrix.min(A);
            Matrix W = Matrix.columnMatch(A,R);
            Stack result = new Stack();
            result.push(R);
            result.push(W);
            return result;
        }
    }
}
class Max extends Builtin{
    Matrix operate(Matrix A)throws MinimatException{
        return Matrix.max(A);
    }
    Stack execute(Stack stack,int nargsin, int nargsout)
     throws MinimatException{
        if(nargsout!=2)
            return super.execute(stack,nargsin,nargsout);
        else{
            Matrix A = (Matrix)stack.pop();
            Matrix R = Matrix.max(A);
            Matrix W = Matrix.columnMatch(A,R);
            Stack result = new Stack();
            result.push(R);
            result.push(W);
            return result;
        }
    }
}
class Exp extends Builtin{}
class LUd extends Builtin{}
class Det extends Builtin{}
class Eig extends Builtin{}
class SVd extends Builtin{}
class QRd extends Builtin{}
