
/*
   SEE WARNING BELOW!

   Builtin functions are designed to behave like interpreted functions.
   There are methods Function.operate() and Function.execute() which
   parallel the methods Builtin.operate() and Builtin.execute() below.
   Both kind of functions are not loaded until invoked by the user.
   Interpreted functions are defined in .m text files on the disk
   and when invoked are compiled by Function.compile; the interpreter
   code is kept in the table Global.functions. Builtin functions are
   contained in container classes which extend Builtin. They are
   constructed when first invoked by the user and stored in the array
   Builtin.functions[] below. If invoked a second time, the corresponding
   entry in the array will be not null and a new instance is not
   constructed. Each container class overrides the operate and execute
   methods it implements.

   Arguments in Minimat are passed by value: in a call B=f(A)
   the function f may change its input without fear of changing
   the value of the matrix A. This is to hold for both Builtin
   functions and interpreted functions. For calls via operate() to
   builtin functions,the responsibility for making the copy is left
   to the interpreter (see case OPn:); other calls pass the execution
   stack and the copy is made by execute() (for builtin functions)
   or bind() (for interpreted functions)

   The names of the builtin functions are kept in the array name[]
   below. Each sub class overrides one or  more of the four functions
   operate() ... operate(Matrix,Matrix,Matrix).    The k-th function
   overides those versions of operate which accept between minArgsin[k]
   and maxArgsin[k] arguments. An error is generated unless the function
   invocation satisfies

     minArgsin[k] <= argsin <= maxArgsin[k] &&
     minArgsout[k] <= argsout <= maxArgsout[k]

   The naming convention for the container classes is as follows:
   Functions which are performed elementwise end in 'E' or 'N';
   the 'N' indicates a function which is `Numeric' and not defined
   over fields more general than the complex numbers. The other
   classes which are handled by BuildBuiltin end in 'M'. Those
   which do not end in a capital letter are defined in
   BuiltinTail.java.

   To facilitate modifying the program, parts of Builtin.java
   are automatically generated by the program BuildBuiltin.
   This prevents errors in the correspondence between the six
   arrays name[],...,function[k]. The first part of Builtin.java is
   copied by BuildBuiltin from the file BuiltinHead.java.  BuildBuiltin
   generates the middle part and then copies the rest from the file
   BuiltinTail.java.

   WARNING: The current version of BuildBuiltin overwrites Builtin.java.


   ToDo: Most of the operations are performed elementwise so that the
   operate method consists of a double loop with a scalar method
   invoked in the body. One could save a bit of space by making the
   the double loop the default and invoking an operate method in
   a container class for the scalar method.
*/


import java.lang.Math;
import java.util.*;

  class Builtin{


    private static final int NOTFOUND = Global.NOTFOUND;

    static boolean find(String id){
        return indexOf(id)!= NOTFOUND;
    }

    private static int indexOf(String id){
        for(int i=0; i<name.length; i++){
            if(id.equals(name[i])) return i;
        }
        return NOTFOUND;
    }

    static Builtin getBuiltin(String id, int nin, int nout)
        throws MinimatException{
        int index = indexOf(id);
        if(index == NOTFOUND) error("Builtin Function '"+id+"' not found");
        if(nin<minArgsin[index]) error("Too few inputs for builtin Function '"+id+"'");
        if(maxArgsin[index]<nin) error("Too many inputs for builtin Function '"+id+"'");
        if(nout<minArgsout[index]) error("Too few outputs for builtin Function '"+id+"'");
        if(maxArgsout[index]<nout) error("Too many outputs for builtin Function '"+id+"'");
        if(functions[index]==null) create(index);
        assert(functions[index]!=null, "getBuiltIn");
        return functions[index];
    }


    String nameOf(){
        for(int i=0; i<functions.length; i++){
            if(this==functions[i])
                return name[i];
        }
        return "Unknown";
    }

    static String dump(Builtin b){
        for(int i=0; i<functions.length; i++){
            if(b==functions[i])
                return name[i];
        }
        return "unknown";
    }


    static String dump(){
        String str="***** Active Buitin Functions *********\n";
        for(int i=0; i<functions.length; i++){
            if(functions[i]!=null)
                str += name[i]+" " + functions[i]+"\n";
        }
        return str;
    }

    private static String unimplemented= "unimplemented function: ";

    Matrix operate()
        throws MinimatException{
        throw new MinimatException(unimplemented+nameOf());
    }

    Matrix operate(Matrix A)
        throws MinimatException{
        throw new MinimatException(unimplemented+nameOf());
    }

    Matrix operate(Matrix A, Matrix B)
        throws MinimatException{
        throw new MinimatException(unimplemented+nameOf());
    }

    Matrix operate(Matrix A, Matrix B, Matrix C)
        throws MinimatException{
        throw new MinimatException(unimplemented+nameOf());
    }


    /* The default method handles stupid invocations like
       [] = f(..) or [Y]=f(...).
    */
    Stack execute(Stack stack,int nargsin, int nargsout)
        throws MinimatException{
        Matrix A=null;Matrix B=null;Matrix C=null;
        Matrix Y;
        switch(nargsin){
            case 0:
                Y= operate();
            break;
            case 1:
                A=new Matrix((Matrix)stack.pop());
                Y= operate(A);
            break;
            case 2:
                B=new Matrix((Matrix)stack.pop());
                A=new Matrix((Matrix)stack.pop());
                Y= operate(A,B);
            break;
            case 3:
                C=new Matrix((Matrix)stack.pop());
                B=new Matrix((Matrix)stack.pop());
                A=new Matrix((Matrix)stack.pop());
                Y= operate(A,B,C);
            break;
            default:throw new MinimatException("unimplemented function");
        }
        Stack result = new Stack();
        if(nargsout == 1) result.push(Y);
        else
        if(nargsout!=0)
        throw new MinimatException("unimplemented function");
        return result;
    }

    static boolean insertBrackets(String id){
    // return true if id is name of a function (like diag)
    // which can be called in either of the forms
    //   id([a b ... c]) or id(a b ... c).

	return
	    id.equalsIgnoreCase("diag") ||
		id.equalsIgnoreCase("max")  ||
		id.equalsIgnoreCase("min")  ||
		id.equalsIgnoreCase("any")  ||
		id.equalsIgnoreCase("all")  ||
		id.equalsIgnoreCase("sum");
    }


    private static void assert(boolean condition, String where){
        Global.assert(condition, "Builtin.java (or BuiltinHead.java): "+where);
    }


     static void error(String msg) throws MinimatException{
        throw new MinimatException(msg);
    }


// *** Add text generated by BuildBuiltin.java

