/*
   The interpreter is designed to treat Builtin functions and
   interprested functions in much the same way.
   The methods Function.operate() and Function.execute() below
   parallel the methods Builtin.operate() and Builtin.execute().
   An important difference is that a Builtin function accepts
   at most 3 inputs (hence four methods Builtin.operate() through
   Builtin.operate(Matrix,Matrix,Matrix)) whereas an interpreted
   function accepts a variable number of inputs (hence one method
   Function.operate(Stack stack, int nargsin, int nargsout))

   Arguments in Minimat are passed by value: in a call B = f(A)
   the function f may change its input without fear of changing
   the value of the variable A. This is to hold for both builtin
   functions and interpreted functions. For calls via operate to
   builtin functions,the responsibility for making the copy is left
   to the interpreter (see case OPn:); other calls pass the execution
   stack and the copy is made by execute() (for Builtin functions)
   or bind() (for interpreted functions)

   Interpreted functions are defined in .m text files on the disk
   and when invoked are compiled by Function.compile; the interpreter
   code is kept in the table Global.functions. If an entry for an
   interpreted function is found in Global.functions, the function is
   not recompiled.

   Function.compile also compiles direct uder input but that input
   is executed by Global.execute, not Function.execute.

*/

import java.util.*;
import java.io.*;

class Function{

    String name; // !=  null if compiling m function

    int maxNargin, maxNargout; // 0,0 for script, user statement

    Vector code;         // code being produced (Instruction)
    TokenVector token;   // Token's being compiled
    Vector variables;    // local variables for this function

    /* Invariant relation: variables.elementAt(k).offset = k
       At execution time variables will be bound to values:
       (Matrix)values.elementAt(k) is the value of the variable
       named ((Variable)variables.elementAt(k)).name.
       if variables==Global.variables then  values==Global.values
       so we add an empty matrix to Global.values whenever we add a
       variable to variables. (See addVariable() below.) If we are compiling
       an interpreted function, it will be invoked via operate() or execute()
       below. These call bind() which is supposed to arrange the correct binding.
    */

    private int tkx=0; // index into token,
    private Token the, next;

    private static final int NOTFOUND = Global.NOTFOUND;

    private void advance(){
        the=next;
	    tkx++;
	    if(tkx<token.size())
	        next = (Token)token.elementAt(tkx);
    }

    private void reset(int tkx1, int codex){// used by nullAssignment
        tkx=tkx1;
        code.setSize(codex);
        next = (Token)token.elementAt(tkx);
    }


    /*
       The methods Function.operate and Function.execute are called
       by Interpreter.interpret() to run interpreted (i.e. .m) functions.
       Direct user input is handled by the method Global.execute().
    */

    Matrix operate(Stack stack, int nargsin)
      throws MinimatException{
        Vector values = bind(stack, nargsin, 1);
        Interpreter proc = new Interpreter(nargsin,1,this,values);;
        proc.interpret();
        return proc.simpleResult();
    }

    Stack execute(Stack stack,int nargsin, int nargsout)
      throws MinimatException{
        Vector values = bind(stack, nargsin, nargsout);
        Interpreter proc = new Interpreter(nargsin,nargsout,this,values);
        proc.interpret();
        return proc.multipleResult();
    }

    /*
        Function.bind insures that the correspondence between
        Interpreter.function.variables and Interpreter.values
        is correct. Note that bind makes a copy of each value
        it passes.
    */
    private Vector bind(Stack stack,int nargsin, int nargsout)
      throws MinimatException{
        assert(nargsin<=maxNargin && nargsout<=maxNargout ,"execute");
        Vector values = new Vector(variables.size());
        for(int i=0; i<variables.size(); i++)
            values.addElement(Matrix.empty);
        for(int i=nargsin-1; i>=0;i--)
            values.setElementAt(new Matrix((Matrix)stack.pop()),maxNargout+i);
        return values;
    }

    Function(Vector variables, TokenVector token)throws MinimatException{

    this.token = token;

    this.name=null; maxNargin=0; maxNargout=0;
    // will be changed if  isMfunction

    if(token.isMfunction()) Global.addFunction(this);
    // recursive calls will now find the function


    if(variables==null || token.isMfunction())
        this.variables = new Vector();
    else
        this.variables = variables;

    tkx=0; the=null; next = (Token)token.elementAt(tkx);

    code = new Vector();
    compile();
    }//Function()

    String dump(){
        String str = "\n****** Variables \n";
        for(int x=0; x<variables.size(); x++){
            Variable v=(Variable)variables.elementAt(x);
            str += v.name+"("+x+")\n";
        }
        str +="\n****** Instructions \n";
        for(int x=0; x<code.size(); x++){
            if(x<10)  str +="  "+x+"   "; else
            if(x<100) str += " "+x+"   "; else
                      str +=  ""+x+"   ";
            Instruction ins = (Instruction) code.elementAt(x);
            str +=  ins.dump()+"\n";
        }
        return str;
    }

    private void compile()throws MinimatException{
    // if we find an error in an mfunction we
    // must discard all compiled mfunctions

	   while (have(Token.EOL)) ; //Do_Nothing
	   if(next.symbol==Token.FUNCTIONW)
	        functionDefinition();
	    else{
	        statements();
	        mustBe(Token.EOT, "end of file");
	        emit(Code.STOP);
	    }
	    verify();
    }


    private void functionDefinition()throws MinimatException{
        mustBe(Token.FUNCTIONW,"function");
	    functionHeading();
	    skip();
	    statements();
	    mustBe(Token.EOT, "end of file");
	    emit(Code.RTN);
	}//functionDefinition



    private void functionHeading()throws MinimatException{
	    if (haveFH(Token.LBRACK)) {
		// function [...] = f(...)
		    maxNargout = 0;
		    while(haveFH(Token.NEWID)){
			    maxNargout++;
			    addVariable(the.text);
			    if (next.symbol != Token.RBRACK)  mustBeFH(Token.COLSEP, "\",\"");
		    }
		    if (haveFH(Token.VARID))
		        error("Illegal repeated identifier in output list");
		    mustBe(Token.RBRACK, "\"]\"");
	    }else
	    if  ((haveFH(Token.NEWID)) && next.symbol == Token.GETS ){
 		    // function v = f(...)
			    addVariable(the.text);
		        maxNargout = 1;
	    }else{
	        error("identifier or \"[\" expected");
        }
	    if(maxNargout !=0) mustBeFH(Token.GETS,"\"=\"");
	    mustBeFH(Token.NEWID,"Identifier");

	    name = the.text;
	    if(token.mFile!=null &&
	       token.mFile.fileName!=null &&
	      !token.mFile.fileName.equalsIgnoreCase(name+".m"))
	        error("Function name does not agree with file name");


	    maxNargin = 0;
	    if (haveFH(Token.LPAREN)) {
            Vector list = new Vector();
		    while(haveFH(Token.NEWID) || haveFH(Token.VARID)){
 			    maxNargin++;
 			    for(int x=0; x<list.size(); x++)
 			        if(the.text.equals((String)list.elementAt(x)))
 			            error("Illegal repeated identifier in argument list");
			    if (the.symbol== Token.NEWID){
			        addVariable(the.text);
			        list.addElement(the.text);
			    }else{//the.symbol==VARID; some output is an input
			        Variable v = getVariable(the.text);
			        String str = "$"+the.text;// illegal name
			        addVariable(str);
			        Variable w = getVariable(str);
			        emit(Code.LDM,w);
			        emit(Code.STM,v);
			    }// if
			    if (next.symbol != Token.RPAREN)
			        mustBeFH(Token.ARGSEP,"\",\"");
		    }//while
            mustBeFH(Token.RPAREN,"\")\"");
	    }//if
    }// functionHeading

    private void statements() throws MinimatException{
	    skip();
	    while (statementSymbol(next.symbol)){
		    statement();
		    if (have(Token.SEMICOLON) || have(Token.COMMA)) ;  //Do_Nothing
		    while (have(Token.EOL)) ;   // Do_Nothing
	    }
    }//statements()



    private void statement()throws MinimatException{
    // called Statement must advance()
	 skip();
	 switch (next.symbol){
		case Token.FORW:    forStatement();
		break;
		case Token.WHILEW:  whileStatement();
		break;
		case Token.BREAKW:  breakStatement();
		break;
		case Token.IFW:     ifStatement();
		break;
		case Token.RETURNW: returnStatement();
		break;
		case Token.FORMATW: formatStatement();
		break;
		case Token.DISPW:   displayStatement();
		break;
		case Token.LBRACK:
		    if (!nullAssignment())	// [....] =
		        multipleAssignment();
		break;
		case Token.VARID:	case Token.BFNID:
		case Token.IFNID:	case Token.NEWID:
			if (!nullAssignment())
			    simpleAssignment();
		break;
		default: nullAssignment();
	 }//switch
    }//statement()

    private void forStatement()throws MinimatException{
	    mustBe(Token.FORW, "for");
	    if (have(Token.NEWID)||have(Token.IFNID)||have(Token.BFNID))
	        addVariable(the.text);
	    else
    	if (have(Token.VARID))
    	    ; // do_Nothing
		else
		    error("Identifier expected");
		Variable v  = getVariable(the.text);
	    mustBe(Token.GETS,"\"=\"");
	    expression();
	    emit(Code.BFOR, v);   // creates forStack entry
	    int batx = code.size();  // Code[bat]  is BAT ..
	    Instruction bat = new Instruction(Code.BAT,null,the);
	    emit(bat);    // do next JMP instruction if count<limit
	    statements();
	    if (next.symbol == Token.EOT)
	        error("for without end");
	    else
		    mustBe(Token.ENDW, "end");
	    emit(Code.JMP,batx); 	//to BAT instruction
	    int end = code.size();
	    bat.arg = new Integer(end);
	    for(int x=batx;x<end;x++){// fix up BRK instructions
	        Instruction ins = (Instruction)code.elementAt(x);
		    if(ins.opCode == Code.BRK && ins.arg == null)
	 		    ins.arg = bat.arg;
	    }
	    emit(Code.EFOR);  		//decrements forStack
	}// forStatement

    private void whileStatement()throws MinimatException{
	    int fix, start;

	    mustBe(Token.WHILEW,"while" );
	    start = code.size();
	    expression();
	    Instruction tst = new Instruction(Code.TST,null,the);
	    emit(tst);
	    statements();
	    if (next.symbol == Token.EOT)
	        error("while without end");
	    else
		    mustBe(Token.ENDW,"end");
	    emit(Code.JMP,start);
	    int end = code.size();
	    tst.arg = new Integer(end);
	    for(int x=start;x<end;x++){// fix up BRK instructions
	        Instruction ins = (Instruction)code.elementAt(x);
		    if(ins.opCode == Code.BRK && ins.arg == null)
	 		    ins.arg = tst.arg;
	    }
    }// whileStatement

    private void breakStatement()throws MinimatException{
	    mustBe(Token.BREAKW,"break;");
		emit(Code.BRK); // fix in forStatement or whileStatement
	}//breakStatement

    private void ifStatement()throws MinimatException{
        Instruction test;
	    mustBe(Token.IFW, "if");
	    expression();
	    test = new Instruction(Code.TST,null,the);
	    emit(test);
	    statements();
	    skip();
	    int start = code.size();
	    while(have(Token.ELSEIFW)){
	        emit(Code.JMP);
	        test.arg = new Integer(code.size());
		    expression();
	        test = new Instruction(Code.TST,null,the);
	        emit(test);
	        statements();
	        skip();
	    }
	    if (have(Token.ELSEW)){
	        emit(Code.JMP);
	        test.arg = new Integer(code.size());
	        statements();
	        skip();
	    }
	    else{
	        test.arg = new Integer(code.size());
	        skip();
	    }
		mustBe(Token.ENDW,"end");
		Integer end = new Integer(code.size());
	    for(int x=start; x<code.size(); x++){
	        Instruction fixup = (Instruction)code.elementAt(x);
	        if(fixup.opCode==Code.JMP && fixup.arg==null)
	            fixup.arg = end;
	    }
    }// ifStatement

    private void returnStatement()throws MinimatException{
	    mustBe(Token.RETURNW, "return");
		emit(Code.RTN);
    }// returnStatement

    private void formatStatement()throws MinimatException{
	    mustBe(Token.FORMATW, "format");
	    expression();
	    emit(Code.FRMT);
    }//formatStatement

    private void displayStatement()throws MinimatException{
	    mustBe(Token.DISPW,"disp");
	    mustBe(Token.PSTRING,"string");
	    emit(Code.OTPT,the.text);
    }//displayStatement

    private void simpleAssignment()throws MinimatException{
        Instruction ins = target(false);
	    mustBe(Token.GETS, "=");
	    expression();
	    emit(ins);
	    if (!have(Token.SEMICOLON)) emit(Code.PRNT,ins.arg);
    }// simpleAssignment



    private boolean nullAssignment()throws MinimatException{
        int tkx1 = tkx;
        int codex = code.size();
        try{
	        expression();
	    }catch(MinimatException ex){//try again
	        reset(tkx1,codex);
	        return false;
	    }
	    if(have(Token.GETS)){ // not a null assignment, try again
	        reset(tkx1,codex);
	        return false;
	    }else{
	        if(getVariable("ans")==null) addVariable("ans");
	        Variable v  = getVariable("ans");
	        emit(Code.STM,v);
	        if (!have(Token.SEMICOLON)) emit(Code.PRNT,v);
	        return true;
	    }
    } // nullAssignment



    private void multipleAssignment()throws MinimatException{
        Stack storeInstruction=new Stack();
	    mustBe(Token.LBRACK,"[");
	    if (!have(Token.RBRACK)){
	        storeInstruction.push(target(true));
	        while (have(Token.COLSEP))
	            storeInstruction.push(target(true));
	    }// if
	    mustBe(Token.RBRACK,"]");
	    mustBe(Token.GETS, "=");
	    if (have(Token.IFNID))
		    multipleInterpretedFunctionCall(storeInstruction.size());
		else
		if (have(Token.BFNID))
 		    multipleBuiltinFunctionCall(storeInstruction.size());
		else
		    error("Unrecognized function call "+next.text);
	    for (int k=storeInstruction.size()-1; k>=0; k--){
			 emit((Instruction)storeInstruction.elementAt(k));
		}
		//emit(Code.FRF);

	    if(!have(Token.SEMICOLON))
	    for (int k=0; k<storeInstruction.size(); k++){
		  Instruction ins = (Instruction)storeInstruction.elementAt(k);
		  emit(Code.PRNT,ins.arg);
		}
     }//multipleAssignment


    private void multipleInterpretedFunctionCall(int nargsout) throws MinimatException{
	    Function func = Global.getFunction(the.text);
	    assert(func != null,"multipleInterpretedFunctionCall");
	    int nargsin = arguments();
	    if (nargsin > func.maxNargin)
		error("Too many inputs to interpreted function "+func.name);
	    else
	    if (nargsout > func.maxNargout)
	    error("Too many outputs to interpreted function "+func.name);
	    else{
		    emit(Code.MICALL,func);
	        emit(Code.SETI,nargsin);
	        emit(Code.SETO,nargsout);
	    }
    }

    private void multipleBuiltinFunctionCall(int nargsout) throws MinimatException{
        String  id=the.text;
	    int nargsin = arguments();
	    if(Builtin.insertBrackets(id)){
	        // convert diag(...) to diag([...])
		    while(nargsin>1){
		        Builtin op = Builtin.getBuiltin(",",2,1);
			    emit(Code.OP2,op);
			    nargsin--;
			}
		}
	    Builtin arg = Builtin.getBuiltin(id,nargsin,nargsout);
		emit(Code.MBCALL,arg);
	    emit(Code.SETI,nargsin);
	    emit(Code.SETO,nargsout);
	}//builtinFunction





    private void expression()throws MinimatException{
        expression(0);
    }
    private void expression(int precedence)throws MinimatException{
	    switch(precedence){

	    case 0: skip();
	        expression(1);
 	        while (have(Token.LOGOP)){
		        String opsy=the.text;
		        expression(1);
   		        operation(2,opsy);
   		    }
        break;
	    case 1:{
	        int k = 0;
    	    while (have(Token.NEGOP)) k++;
	        expression(2);
	        while (k>0) {
		        operation(1,"~");
		        k--;
 	        }
 	    }break;
	    case 2:
	        expression(3);
	        if (have(Token.RELOP) ){
		        String opsy = the.text;
		        expression(3);
		        operation(2,opsy);
	        }
	    break;
	    case 3:
	        expression(4);
	        if (have(Token.COLON)){
		        expression(4);
		        if (have(Token.COLON)){
			        expression(4);
		            Builtin op = Builtin.getBuiltin(":",2,1);
		            emit(Code.OP3,op);
		        }else{
		            Builtin op = Builtin.getBuiltin(":",2,1);
			        emit(Code.OP2,op);
		        }
	        }
	    break;
	    case 4:{
	        char op = have(Token.ADDOP) ? the.text.charAt(0):'+';
	        expression(5);
	        if (op == '-') operation(1,"-");
	        while (have(Token.ADDOP) ){
		        String opsy = the.text;
		        expression(5);
		        operation(2,opsy);
	        }
	    }break;
	    case 5:
	        expression(6);
            while (have(Token.MULOP) ){
		        String opsy = the.text;
		        expression(6);
		        operation(2,opsy);
		    }
	    break;
        case 6:
	        simpleExpression();
	        while (have(Token.POSTOP)) operation(1,the.text);
	        if (have(Token.EXPOP)){
		        String opsy = the.text;
		        simpleExpression();
		        operation(2,opsy);
	        }
	        while (have(Token.POSTOP)) operation(1,the.text);
 	    break;
 	    default: assert(false,"expression");
        }// switch
    }// expression

    private void simpleExpression()throws MinimatException{
	    int  nargsin;
	    if (have(Token.NUMBER)){
	        try{
		    emit(Code.VAL, new Double(the.text));
		    }
		    catch(NumberFormatException e){
		        error("Bad Number: "+e.getMessage());
		    }
	    }else
	    if (have(Token.LPAREN)){
		    expression();
		    mustBe(Token.RPAREN,")");
	    }else
	    if (have(Token.LBRACK))
		    constructor();
	    else
	    if (have(Token.LCURLY))
		    bracketExpression();
	    else
	    if (have(Token.INPUTW))
		    inputExpression();
	    else
	    if (have(Token.NARGINW)){
		    emit(Code.GETI);
	    }else
	    if (have(Token.NARGOUTW)){
		    emit(Code.GETO);
	    }else
	    if (have(Token.VARID)){
		    Variable v  = getVariable(the.text);
		    emit(subscripts(),v);
	    }else
	    if (have(Token.IFNID)){
		    simpleInterpretedFunctionCall();
		    //emit(Code.FRF);
	    }else
	    if (have(Token.BFNID)){
		    simpleBuiltinFunctionCall();
        }else
	    if (have(Token.NEWID))
		    error("Unrecognized identifier: "+the.text);
	    else
	    if (next.symbol == Token.EOL || next.symbol == Token.EOT)
		    error("Unexpected end of input");
	    else
		    error("Expression expected");
    }//simpleExpression


    private void operation(int nargsin, String opsy) throws MinimatException{
		Builtin arg = Builtin.getBuiltin(opsy,nargsin,1);
        if(nargsin==1)      emit(Code.OP1,arg);
        else /*nargsin==2*/ emit(Code.OP2,arg);
    }

    private void simpleInterpretedFunctionCall() throws MinimatException{
	    Function func = Global.getFunction(the.text);
	    assert(func != null,"simpleInterpretedFunctionCall");
	    int nargsin = arguments();
	    if (nargsin <= func.maxNargin){
		    emit(Code.SICALL,func);
	        emit(Code.SETI,nargsin);
	    }
	    else
		    error("Too many inputs to interpreted function "+func.name);
    }



    private void simpleBuiltinFunctionCall() throws MinimatException{
        String  id=the.text;
	    int nargsin = arguments();
	    if(Builtin.insertBrackets(id)){
	        // convert diag(...) to diag([...])
		    while(nargsin>1){
		        Builtin op = Builtin.getBuiltin(",",2,1);
			    emit(Code.OP2,op);
			    nargsin--;
			}
		}
		Builtin arg = Builtin.getBuiltin(id, nargsin,1);
		switch(nargsin){
		 case 0: emit(Code.OP0,arg); break;
		 case 1: emit(Code.OP1,arg); break;
		 case 2: emit(Code.OP2,arg); break;
		 case 3: emit(Code.OP3,arg); break;
	    }
	}//builtinFunction



    private Instruction target(boolean multiple)throws MinimatException{
	    if (have(Token.NEWID)||have(Token.IFNID)||have(Token.BFNID)) {
	        addVariable(the.text);
		    Variable v = getVariable(the.text);
		    if ((next.symbol== Token.LPAREN)) error("variable expected");
		    if(multiple)
		        return new Instruction(Code.MSTM,v,the);
		    else
		        return new Instruction(Code.STM, v,the);
        }else
  	    if (have(Token.VARID)){
		    Variable v = getVariable(the.text);
		    int op = subscripts()+1;// SX=LX+1
		    if(multiple) op += 1;   // MSX=SX+1
		    return new Instruction(op,v,the);
	    }else{
		    error("identifier expected");
		    return null;
	    }
    }// target

    private int subscripts()throws MinimatException{
	    int op;
	    if  (!have(Token.LPAREN))  op = Code.LDM;
   	    else
   	    if (have(Token.COLON)){// (:) or (:,*)
   		    if (have(Token.RPAREN)) op = Code.LDC;
        	else{ //(:,*)
			    mustBe(Token.ARGSEP, "\",\"");
			    if (have(Token.COLON)) op  = Code.LDM;
			    else {expression(); op = Code.LDCS;}
			    mustBe(Token.RPAREN,"\")\"");
		    }
	    }else{// (*) or (*,:) or (*,*)
		    expression();
		    if (have(Token.RPAREN)) op = Code.LDS;
		    else{ // (*,:) or (*,*)
			    mustBe(Token.ARGSEP,"\",\"");
			    if (have(Token.COLON)) op = Code.LDSC;
			    else { // (*,*)
			        expression();   op = Code.LDSS;
			    }
			    mustBe(Token.RPAREN, "\")\"");
		    }
	    }// if elseif else
	    return op;
    }// subscripts()




    private int arguments()throws MinimatException{
	    int nin = 0;
	    if (have(Token.LPAREN)){
		  while (expressionSymbol(next.symbol)){
			expression();
			nin++;
			if (next.symbol != Token.RPAREN ) mustBe(Token.ARGSEP,"\",\"");
		  }
		  mustBe(Token.RPAREN,"\")\"");
	    }
	    return nin;
    }// arguments()


    private void constructor()throws MinimatException{
	    //call is: if (have(Token.LBRACK))  constructor();
	    if (have(Token.RBRACK) ){
		    return;
	    }
	    expression();
		Builtin colcat = Builtin.getBuiltin(",",2,1);
		Builtin rowcat = Builtin.getBuiltin(";",2,1);
	    while (have(Token.COLSEP)){
		    expression();
		    emit(Code.OP2,colcat);
	    }
	    while (have(Token.ROWSEP)){
		    expression();
		    while (have(Token.COLSEP)){
    			expression();
			    emit(Code.OP2,colcat);
		    }
		    emit(Code.OP2,rowcat);
	    }
	    mustBe(Token.RBRACK,"\"]\"");
    }//constructor()

    private void inputExpression()throws MinimatException{
	    mustBe(Token.PSTRING,"('");
    	emit(Code.INPT,the.text);
    }

    private void bracketExpression()throws MinimatException{
	    expression();
	    mustBe(Token.ARGSEP, "\",\"");
	    expression();
	    mustBe(Token.RCURLY,"}");
	    Builtin commutator = Builtin.getBuiltin("{}",2,1);
	    emit(Code.OP2,commutator);
    }



    private void addVariable(String name){
        Variable v = new Variable(name, variables.size());
        variables.addElement(v);
        if(variables == Global.variables) // see invariant relation at top of file
            Global.values.addElement(Matrix.empty);
    }

    private Variable getVariable(String str){
        for(int vx=0; vx<variables.size();vx++){
            Variable v=(Variable) variables.elementAt(vx);
            if(str.equals(v.name)) return v;
        }
        return null;
    }


    private void emit(Instruction ins){
        code.addElement(ins);
    }

    private void emit(int opCode,   Object arg){
        emit(new Instruction(opCode,  arg, the));
    }

    private void emit(int opCode){
        emit(opCode, null);
    }

    private void emit(int opCode, int arg){
        emit(opCode, new Integer(arg));
    }


    private boolean statementSymbol(int sy)
    throws MinimatException{// a statement begins with one of these
        look();
        switch (sy ){
        case Token.LPAREN: case Token.LBRACK:
        case Token.NEGOP:  case Token.ADDOP:
        case Token.ID:     case Token.NEWID:
        case Token.VARID:  case Token.BFNID:   case Token.IFNID:
        case Token.NUMBER: case Token.FORMATW:
        case Token.DISPW:  case Token.INPUTW:
        case Token.IFW:    case Token.FORW:    case Token.WHILEW:
        case Token.BREAKW: case Token.RETURNW:
    	    return true;
        default:
     	    return false;
        }
    }//statementSymbol()


    private boolean expressionSymbol(int sy)
    throws MinimatException{// an expression begins with one of these
        look();
        switch( sy ){
	    case Token.LPAREN: case Token.LBRACK:
	    case Token.NEGOP:  case Token.ADDOP:
        case Token.ID:     case Token.NEWID:
        case Token.VARID:  case Token.BFNID: case Token.IFNID:
        case Token.NUMBER: case Token.INPUTW:
    	    return true;
        default:
     	    return false;
	    }
    }//expressionSymbol()


    private void skip() throws MinimatException{
    while (have(Token.EOL) || have(Token.SEMICOLON) || have(Token.COMMA)) ;// Do_Nothing
    }


    private void look()throws MinimatException{
	    if (next.symbol != Token.ID ) return;
	    String str =  next.text;
	    if (getVariable(str) != null)
		    next.symbol=Token.VARID;
	    else
	    if (Global.getFunction(str) != null)
		    next.symbol=Token.IFNID;
		else
	    if (Builtin.find(str))
		    next.symbol=Token.BFNID;
	    else
		    next.symbol=Token.NEWID;
    }



    private boolean have(int sy) throws MinimatException{
	    look();
	    if (sy==next.symbol){
		    advance();
		    return true;
	    }
	    else return false;
    }

    private void mustBe(int sy, String msg)throws MinimatException{
        look();
	    if (sy==next.symbol) advance();
	    else{
	        String str = next.text;
	        if(str==null || str.length()==0) str="";
	        error(next, msg+" expected after "+the.text+str);
	    }
    }

    private void lookFH()throws MinimatException{
	    if (next.symbol != Token.ID ) return;
	    String str =  next.text;
	    if (getVariable(str) != null)
		    next.symbol=Token.VARID;
	    else
		    next.symbol=Token.NEWID;
    }



  // haveFH replaces have in functionHeading
    private boolean haveFH(int sy) throws MinimatException{
	    lookFH();
	    if (sy==next.symbol){
		    advance();
		    return true;
	    }
	    else return false;
    }

  // mustBeFH replaces mustBe in functionHeading
    private void mustBeFH(int sy, String msg)throws MinimatException{
        lookFH();
	    if (sy==next.symbol) advance();
	    else error(next, msg+" expected, not "+next.text);
    }

    private void error(String message)throws MinimatException{
        error(the, message);
    }
    private void error(Token tkn, String message)throws MinimatException{
        if(token.mFile!=null){
            message += "\nError occurred in file "
                    + token.mFile.filePath
                    + "\nat line number "
                    + tkn.lineNumber(token.ambientText);
        }
        throw new MinimatException(tkn,message);
    }

    private static void assert(boolean condition, String message){
        Global.assert(condition, "Function.java: "+message);
    }

//ToDo
/*
   verify() checks the validity of the code.
   It calls Instruction.valid to make sure
   each opCode has an argument of the corresponding
   type and it also checks that the calling sequences.
   i.e. instructions following the opcodes
        Code.SICALL  (simpleInterpretedFunctionCall)
		Code.SBCALL  (simpleBuiltinFunctionCall)
	   	Code.MICALL  (multipleInterpretedFunctionCall)
	   	Code.MBCALL  (multipleBuiltinFunctionCall),
   are correct.
*/
private void verify(){
    for(int i=0; i<code.size(); i++){
        Instruction ins = (Instruction) code.elementAt(i);
        assert(ins.valid(),"verify: "+name+" instruction at "+i+"\n"+ins.dump());
    }
}
/*
    private boolean checkICall(int pc){
        return (pc+2<func.code.size()) &&
  		(((Instruction)func.code.elementAt(pc+1)).opCode==Code.SETI)&&
  		(((Instruction)func.code.elementAt(pc+2)).opCode==Code.SETO);
    }

    private boolean checkBCall(int pc, int nout){
        if(pc+nout >= func.code.size()) return false;
        if(result==null) return false;
        if(result.size() != nout) return false;
        for(int k=1;k<=nout; k++){
            int op = ((Instruction)func.code.elementAt(pc+k)).opCode;
            if(op!=Code.MSTM && op!=Code.MSTC && op!=Code.MSTCS &&
               op!=Code.MSTS && op!=Code.MSTSC && op!=Code.MSTSS )
                return false;
        }
        return true;
    }
    */


//ToDo
/*
   It is possible to determine the size of the execution
   stack in advance. This might lead to a more efficient
   implementation - e.g. a customized version of Stack.
*/
private void calibrate(){
}
}//Function