
import java.lang.Math;
import java.util.Vector;
import java.io.*;

class TokenVector extends Vector{

    /*
       The task of parsing a String text into a Vector of
       tokens is complicated by the fact that the representation
       of a token in Minimat is somewhat context dependent.

       In BMODE (between '[' and ']')   either  ' ' or   ','
       might separate columns and either '\n' or ';' might separate
       rows. The text between  '%' or ".." an '\n' is treated as
       a comment and hence ignored,        but after ".." the next
       '\n' is also ignored. This allows the user to continue logical
       lines on more than one physical line.

       In PMODE (between '(' and ')') the space ' ' has no significance
       and   ',' is used to separate arguments.

       In SMODE (between '\'' and '\'') we are scanning a string.
       The only place a string appears in Minimat is in the context
       disp('string')  or input('string'). Hence "(\'" begins
       SMODE and "\')" ends it.

       Minimat has no string handling facility. A string
       may appear only in one of the two contexts
       disp('string') and input(`string').
    */


    DotMFile mFile;  //    =  null if  keyboard input

    String ambientText;
    private int tx;     // index into ambientText
    private int txmax; // = ambientText.length()

    private final static int MODEMAX = 10;
    private final static int OMODE=0; // Ordinary mode
    private final static int PMODE=1; // Parenthesis mode
    private final static int BMODE=2; // Bracket mode
    private final static int SMODE=3; // String mode
    private int[] modeStack= new int[MODEMAX]; // never empty
    private int mtop; //  modeStack[0..mtop] meaningful

    private final static int ZILCH=0;
    private final static int BLANK=1;
    private final static int ENDLINE=2;
    private int seenByStrongSkip; // One of the above three. Set by StrongSkip
    private boolean found; // Set false by getSym, set true by adjoin, used to exit getSym.


    TokenVector(String ambientText)throws MinimatException{
        super(10, 10);
        this.mFile=null;
        construct(ambientText);
    }


     TokenVector(DotMFile mFile) throws MinimatException{
        super(100,10);
		this.mFile=mFile;
		construct(mFile.getContents());
    }


    private void construct(String ambientText)throws MinimatException{
	    mtop = 0;
	    modeStack[mtop] = OMODE;
	    this.ambientText=ambientText;
	    tx = 0;
	    txmax = ambientText.length();
	    while(tx<txmax) getSym();
	    adjoin(Token.EOT,0);
	 }

    int symbol(int x){
        return ((Token)elementAt(x)).symbol;
    }

    boolean isMfunction(){
        int x=0;
        while(symbol(x)==Token.EOL) x++;
        return (symbol(x)==Token.FUNCTIONW);
    }

    String dump(){
            String str = "";
            for(int x=0;x<elementCount; x++)
                str += ((Token)elementAt(x)).toString(ambientText)+"\n";
            return str;
    }


    private void error(String message) throws MinimatException{
        if(mFile!=null){
            message += "\nBad token found in file "
                    +  mFile.filePath
                    +  "\nat line number "
                    +  ((Token)lastElement()).lineNumber(ambientText);
        }
        throw new MinimatException((Token)lastElement(),message);
    }

    private void getSym()throws MinimatException{
	found = false;
	switch (modeStack[mtop]){
		case OMODE:	seperatorO(); break;
        case BMODE: seperatorB(); break;
        case PMODE:	seperatorP(); break;
        case SMODE: scanPstring();break;
        default: throw(new IllegalArgumentException("JWR-TokenVector.getSym"));
	}
   	if (found || tx>=txmax) return;
    if (isAlpha(ch()))           scanIdentifier(); else
    if (Character.isDigit(ch())) scanNumber();     else
                                 scanSpecial();
    }// getSym

    private char filter(char kh){
        if(' ' <= kh && kh <= '~') return kh;
        if(kh=='\n') return kh;
        return ' ';
    }

    private char ch(){
      if(tx<txmax)
        return filter(ambientText.charAt(tx));
      else
        return '@';
    }

    private char nextch(){
        if(tx+1<txmax)
            return filter(ambientText.charAt(tx+1));
        else
            return '@';
    }

    private void advance(){
        if (tx<txmax) tx++;
    }

    private void adjoin(int sy, int len){
	// Assumes ch() is first char not in token found
	found=true; // exit from getSym
	addElement(new Token(sy, tx-len,tx,ambientText));
    }

    private void comment(int flag){
    	// advance  to end of line or to end of ambientText
      	while (tx < txmax &&  ch() != '\n') advance();
      	if (flag == '.' ) advance();// i.e. ignore '\n'
    }

    private void weakSkip(){// advance to non-blank token
	for(;;){
		if ( tx>= txmax) return;
		switch (ch()){
			case ' ':  advance();  break;
			case '%':  comment('%'); break;
			case '.':
				if (nextch() == '.') comment('.');
				else return;
			break;
			default: return;
		}// switch
	}// for
    }// weakSkip


    private void StrongSkip(){
    //skip all implicit punctuation. Report  what was found
     seenByStrongSkip = ZILCH;
     for(;;){
        if (tx >= txmax)  return;
        switch (ch()) {
        	case ' ':
        		advance();
        		if (seenByStrongSkip != ENDLINE)  seenByStrongSkip = BLANK;
        	break;
        	case '\n':
        		advance();
        		seenByStrongSkip = ENDLINE;
        	break;
        	case '%':
        		comment('%');
        		seenByStrongSkip = ENDLINE;
        	break;
        	case '.': if (nextch() == '.') comment('.'); else return;
			break;
			default: return;
		  }// switch
	}// for
    }// strongSkip



    private void opSkip(){
    // called after an operator is found
	for(;;){
        if (tx >= txmax) return;
        switch (ch()){
        	case ' ' :  advance(); break;
			case '%' :  comment('%'); break;
			case '\n':  advance();  break;
          	case '.' :  if (nextch() == '.') comment('.'); else return; break;
            default: return;
		}// switch
	}//for
    }//opSkip


    private void seperatorO()throws MinimatException{
    // Called in OMODE
	weakSkip();
  	if ( tx >= txmax ) return;
	if (ch() == ',' ){
	    advance();
		adjoin(Token.COMMA, 1);
	}else
	if (ch() == ';' ){
	    advance();
		adjoin(Token.SEMICOLON, 1);
	}else
	if (ch() == '\n' ){
	    advance();
		adjoin(Token.EOL,0); // empty ambientText
    }
    }// seperatorO

    private void seperatorB()throws MinimatException{
    // called in BMODE
	StrongSkip();
  	if ( tx >= txmax ) return;
	if (ch() == ',' ){
	    advance();
		adjoin(Token.COLSEP, 1);
		StrongSkip();
	}
	else
	if (ch() == ';' ){
	    advance();
		adjoin(Token.ROWSEP,1);
		StrongSkip();
     }
     else
     if (ch() == ']' || ch() == ')' ){
		; //Do_Nothing
     }else
     if (seenByStrongSkip == BLANK){
		adjoin(Token.COLSEP, 0);
     }else
     if (seenByStrongSkip == ENDLINE)
		adjoin(Token.ROWSEP, 0);
    }//seperatorB

    private void seperatorP()throws MinimatException{
    // Called in PMODE
	StrongSkip();
  	if ( tx >= txmax ) return;
	if (ch() == ',' ){
	    advance();
		adjoin(Token.ARGSEP, 1);
	}else
	if (ch() == ';' ){
	    advance();
		adjoin(Token.ERR,1);
		error("Use ',' here not ';'");
	}
    }//seperatorP


    private void scanPstring()throws MinimatException{
    // called in SMODE
	int len;
	StrongSkip();
	if ( ch() != '(' ) {
	    advance();
		adjoin(Token.ERR,1);
		error("Left parenthesis  \"(\"  expected ");
		return;
    }
    advance();
	StrongSkip();
    if (ch()  =='\''){
		advance();
		len = 0;
	}else{
	    advance();
		adjoin(Token.ERR,1);
		error("String  (i.e. quote mark \"'\")  expected");
		return;
	}
	while (tx<txmax && ch() != '\n' && ch() != '\'') {advance(); len++;}
	if (tx>=txmax  || ch() == '\n'){
	    advance();
		adjoin(Token.ERR,len);
		error("Unterminated string");
        return;
	}else{ // ch == '\''
		adjoin(Token.PSTRING,len);
		advance();
	}
	StrongSkip();
    if (ch() != ')'){
        advance();
        adjoin(Token.ERR,1);
        error("')' expected");
    	return;
    }
    else{
		if(mtop>0) mtop--; // popMode
		advance();
	}
    }//scanPstring


    private void pushMode(int mode)throws MinimatException{
	    if ( mtop >= MODEMAX ){
		    advance();
		    adjoin(Token.ERR,1); // such nesting is the result of malice
		    error("Nesting too deep");
    	}else{
		    mtop++;
		    modeStack[mtop] = mode;
		    opSkip();
	    }
    }// pushMode

    private void popMode() {if(mtop>0) mtop--;}


    private boolean isAlpha(char kh){
        return ('A'<= kh && kh <='Z') || ('a'<= kh && kh <='z');
    }

    private void scanIdentifier()throws MinimatException{
    // assume isAlpha(ch())
    int  len=0;
    do{advance(); len++;
	}while (tx<txmax && isAlpha(ch()) || Character.isDigit(ch()));
    String str = ambientText.substring(tx-len,tx);

    if(str.equalsIgnoreCase("break"))    adjoin(Token.BREAKW,len); else
    if(str.equalsIgnoreCase("disp"))    {adjoin(Token.DISPW,len); pushMode(SMODE);} else
    if(str.equalsIgnoreCase("else"))     adjoin(Token.ELSEW,len); else
    if(str.equalsIgnoreCase("elseif"))   adjoin(Token.ELSEIFW,len); else
    if(str.equalsIgnoreCase("end"))      adjoin(Token.ENDW,len); else
    if(str.equalsIgnoreCase("for"))      adjoin(Token.FORW,len); else
    if(str.equalsIgnoreCase("format"))   adjoin(Token.FORMATW,len); else
    if(str.equalsIgnoreCase("function")) adjoin(Token.FUNCTIONW,len); else
    if(str.equalsIgnoreCase("if"))       adjoin(Token.IFW,len); else
    if(str.equalsIgnoreCase("input"))   {adjoin(Token.INPUTW,len); pushMode(SMODE);} else
    if(str.equalsIgnoreCase("nargin"))   adjoin(Token.NARGINW,len); else
    if(str.equalsIgnoreCase("nargout"))  adjoin(Token.NARGOUTW,len); else
    if(str.equalsIgnoreCase("return"))   adjoin(Token.RETURNW,len); else
    if(str.equalsIgnoreCase("while"))    adjoin(Token.WHILEW,len); else
    adjoin(Token.ID,len);
 }// scanIdentifier

  private void scanNumber()throws MinimatException{
	int len=0;
    while (tx<txmax && Character.isDigit(ch())){advance(); len++;}
    if (ch() == '.') {//decimal
        advance(); len++;
		while (tx<txmax && Character.isDigit(ch())){advance(); len++;}
	}
    if (Character.toUpperCase(ch()) == 'E') {// exponent
		advance(); len++;
		if (ch()== '+' || ch() == '-') {advance(); len++;}
		if(!Character.isDigit(ch())) {
		    adjoin(Token.ERR,1);
 	    error("Missing exponent");
 	    }
    	while (tx<txmax && Character.isDigit(ch())){advance(); len++;}
	}
    adjoin(Token.NUMBER,len);
  	if (isAlpha(ch()) || Character.isDigit(ch())){
 	  adjoin(Token.ERR,1);
 	  error("Number next to letter");
 	}
}//scanNumber

 private void scanSpecial()throws MinimatException{
	switch (ch()){
        case '(': 	advance(); adjoin(Token.LPAREN,1); pushMode(PMODE); break;
	   	case '{':	advance(); adjoin(Token.LCURLY,1); pushMode(PMODE); break;
 		case '[': 	advance(); adjoin(Token.LBRACK,1); pushMode(BMODE); break;
		case ')':	advance(); adjoin(Token.RPAREN,1); popMode(); break;
		case '}':	advance(); adjoin(Token.RCURLY,1); popMode(); break;
		case ']':	advance(); adjoin(Token.RBRACK,1); popMode(); break;
		case '^':	advance(); adjoin(Token.EXPOP,1);  opSkip();break;
		case '+':	advance(); adjoin(Token.ADDOP,1);  opSkip();break;
		case '-':	advance(); adjoin(Token.ADDOP,1);  opSkip();break;
		case '*':	advance(); adjoin(Token.MULOP,1);  opSkip();break;
		case '/':	advance(); adjoin(Token.MULOP,1);  opSkip();break;
		case '=':
			if (nextch() == '=') {
				advance(); advance(); adjoin(Token.RELOP,2); opSkip();
             }else{
             	advance(); adjoin(Token.GETS,1); opSkip();
             }
        break;
		case '&':	advance(); adjoin(Token.LOGOP,1);  opSkip();break;
		case '|':	advance(); adjoin(Token.LOGOP,1);  opSkip();break;
		case '~':
        	if (nextch()  == '=') {
        		advance(); advance(); adjoin(Token.RELOP,2); opSkip();
       		}else{
       			advance(); adjoin(Token.NEGOP,1); opSkip();
            }
		break;
        case '<':
        	if (nextch() == '=') {
        		advance(); advance(); adjoin(Token.RELOP,2); opSkip();
         	}else
            if (nextch() == '>') {
            	advance(); advance(); adjoin(Token.RELOP,2); opSkip();
             }else{
             	advance(); adjoin(Token.RELOP,1); opSkip();
             }
		break;
        case '>':
			if  (nextch() == '=') {
				advance();  advance(); adjoin(Token.RELOP,2); opSkip();
			}else{
				advance(); adjoin(Token.RELOP,1); opSkip();
			}
		break;
        case ':':	advance(); adjoin(Token.COLON,1);  opSkip();break;
        case '\'':	advance(); adjoin(Token.POSTOP,1); opSkip();break;
        case '.':
        	if  (Character.isDigit(nextch()))
            	scanNumber();
 			else
 			if  ( nextch() == '*'){
 				advance(); advance(); adjoin(Token.MULOP,2); opSkip();
            }else
            if  ( nextch() == '/'){
            	advance(); advance(); adjoin(Token.MULOP,2); opSkip();
            }else
            if  ( nextch() == '\'') {
            	advance(); advance(); adjoin(Token.POSTOP,2); opSkip();
			}else
            if ( nextch() == '^'){
            	advance(); advance(); adjoin(Token.EXPOP,2); opSkip();
            }else{
                advance(); adjoin(Token.ERR,1);
                error("The character " +
                ambientText.substring(tx-1,tx) +
                " (ascii code = " +
                (int)ambientText.charAt(tx-1)  +
                ") should not follow '.'");
            }
		break;
        default:
        	advance(); adjoin(Token.ERR,1);
        	error("The character " +
        	      ambientText.substring(tx-1,tx) +
        	     " (ascii code = " +
        	      (int)ambientText.charAt(tx-1)  +
        	      ") is an illegal character");
      }//switch
    }//scanSpecial



	final static int BALMAX = 10;
    int balanced(){
	int bal=0;
   	int[] stack=new int[BALMAX];
   	int kx; int tc=0; int flag;

	for (kx=0; kx < elementCount;  kx++){
		switch (symbol(kx)){
		case Token.LPAREN: flag= 1;  tc=Token.RPAREN; break;
		case Token.LBRACK: flag= 1;  tc=Token.RBRACK; break;
		case Token.LCURLY: flag= 1;  tc=Token.RCURLY; break;
		case Token.IFW:    flag= 1;  tc=Token.ENDW;   break;
		case Token.FORW:   flag= 1;  tc=Token.ENDW;   break;
		case Token.WHILEW: flag= 1;  tc=Token.ENDW;   break;
		case Token.RPAREN: flag= -1;  break;
		case Token.RBRACK: flag= -1;  break;
		case Token.RCURLY: flag= -1;  break;
		case Token.ENDW:   flag= -1;  break;
		case Token.ERR:
		case Token.EOT: return Token.EOT;
	  	default: 	   flag=0;
		}//switch
		if (flag==1){
			bal++;
			if (bal==BALMAX) return Token.EOT;
		 	else stack[bal]=tc;
		} else	if (flag==-1){
			if (bal== 0) return Token.EOT;
			if (symbol(kx)!=stack[bal]) return Token.EOT;
			bal--;
		}
	}// for
	if (elementCount==0) // empty line
  		return Token.EOT;
  	else
  	if (bal>0)
  		return stack[bal];
  	else {
  		if (Token.POSTOP<symbol(kx-1) && symbol(kx-1)<=Token.GETS)
  			return Token.EOL;
  		else
  			return Token.EOT;
  	}
	}//balanced


    private static void assert(boolean condition, String where){
        Global.assert(condition, "JWR-TokenVector."+where);
    }


}// TokenVector
